import warnings
import numpy as np

# Scipy
try:
    import scipy.linalg as spa
    from scipy.spatial import ConvexHull
    from scipy.linalg import block_diag
except:
    warnings.warn("You don't have scipy package installed. You may get error while using some feautures.")

# Pydrake
try:
    import pydrake.solvers.mathematicalprogram as MP
    import pydrake.solvers.gurobi as Gurobi_drake
    import pydrake.solvers.osqp as OSQP_drake
    # use Gurobi solver
    global gurobi_solver,OSQP_solver, license
    gurobi_solver=Gurobi_drake.GurobiSolver()
    license = gurobi_solver.AcquireLicense()
    OSQP_solver=OSQP_drake.OsqpSolver()
except:
    warnings.warn("You don't have pydrake installed properly. Methods that rely on optimization may fail.")
    

# Pypolycontain
try:
    import pypolycontain as pp
except:
    warnings.warn("You don't have pypolycontain properly installed. Can not import objects")


"""
Optimization-based Operations:
"""  


def affine_map( T, P, t=None ):
    """
    Returns the affine map of a polytope.
    """        
    if type(t)==type(None):
            t=np.zeros((P.n,1))
    if P.type=='AH_polytope':
        return pp.AH_polytope(t=t+np.dot(T,P.t),T=np.dot(T,P.T),P=P.P)
    elif P.type=='zonotope':
        return pp.zonotope(x=t+np.dot(T,P.x),G=np.dot(T,P.G))
    elif P.type=="H_polytope":
        if T.shape[0]>=T.shape[1]:
            Tinv=np.linalg.pinv(T)
            H=np.dot(P.H,Tinv)
            return pp.H_polytope(H=H,h=P.h+np.dot(H,t))
        else:
            Q=pp.to_AH_polytope(P)
            return affine_map( T, Q, t )
    else:
        return ValueError('Polytope type: ',P.type," Not recognized")

def translate( t, P ):
    """
    Shifts the polytope by t vector
    """ 
    assert t.shape[0]==P.n # Dimension match       
    if P.type=='AH_polytope':
        return pp.AH_polytope(t=t+P.t,T=P.T,P=P.P)
    elif P.type=='zonotope':
        return pp.zonotope(x=t+P.x,G=P.G)
    elif P.type=="H_polytope":
        return pp.H_polytope(H=P.H,h=P.h+np.dot(P.H,t))
    else:
        return ValueError('Polytope type: ',P.type," Not recognized")
              
def point_membership(Q,x,tol=10**-5,solver="gurobi"):
    if type(Q).__name__=="H_polytope":
        return Q.if_inside(x,tol)
    else:
        Q=pp.to_AH_polytope(Q)
        prog=MP.MathematicalProgram()
        zeta=prog.NewContinuousVariables(Q.P.H.shape[1],1,"zeta")
        prog.AddLinearConstraint(A=Q.P.H,ub=Q.P.h+tol,lb=-np.inf*np.ones((Q.P.h.shape[0],1)),vars=zeta)
        prog.AddLinearEqualityConstraint(Q.T,x-Q.t,zeta)
        if solver=="gurobi":
            result=gurobi_solver.Solve(prog,None,None)
        elif solver=="osqp":
            prog.AddQuadraticCost(np.eye(zeta.shape[0]),np.zeros(zeta.shape),zeta)
            result=OSQP_solver.Solve(prog,None,None)
        else:
            result=MP.Solve(prog)
    return result.is_success()

def check_subset(P1,P2,k=-1):
    """
    Checks if .math..`P1 \subseteq P2` 
    Inputs:
        
    """

def point_membership_fuzzy(Q,x,tol=10**-5,solver="gurobi"):
    """
    Fuzzy membership check. If x contains NaN, the entry is unconstrained
    @param Q: Polytope in R^n
    @param x: n*1 numpy array, may contain NaNs
    @param tol:
    @param solver: solver to use
    @return: boolean of whether x is in Q
    """
    Q=pp.to_AH_polytope(Q)
    prog=MP.MathematicalProgram()
    zeta=prog.NewContinuousVariables(Q.P.H.shape[1],1,"zeta")
    prog.AddLinearConstraint(A=Q.P.H,ub=Q.P.h+tol,lb=-np.inf*np.ones((Q.P.h.shape[0],1)),vars=zeta)
    assert(x.shape[1]==1)
    for i, xi in enumerate(x):
        if not np.isnan(xi):
            prog.AddLinearEqualityConstraint(np.atleast_2d(Q.T[i,:]),(x[i]-Q.t[i]).reshape([-1,1]),zeta)
    if solver=="gurobi":
        result=gurobi_solver.Solve(prog,None,None)
    elif solver=="osqp":
        prog.AddQuadraticCost(np.eye(zeta.shape[0]),np.zeros(zeta.shape),zeta)
        result=OSQP_solver.Solve(prog,None,None)
    else:
        result=MP.Solve(prog)
    return result.is_success()

def check_non_empty(Q,tol=10**-5,solver="gurobi"):
    Q=pp.to_AH_polytope(Q)
    prog=MP.MathematicalProgram()
    zeta=prog.NewContinuousVariables(Q.P.H.shape[1],1,"zeta")
    prog.AddLinearConstraint(A=Q.P.H,ub=Q.P.h+tol,lb=-np.inf*np.ones((Q.P.h.shape[0],1)),vars=zeta)
    if solver=="gurobi":
            result=gurobi_solver.Solve(prog,None,None)
    elif solver=="osqp":
        prog.AddQuadraticCost(np.eye(zeta.shape[0]),np.zeros(zeta.shape),zeta)
        result=OSQP_solver.Solve(prog,None,None)
    else:
        result=MP.Solve(prog)
    return result.is_success()

def directed_Hausdorff_distance(Q1,Q2,ball="infinty_norm",solver="gurobi"):
    r"""
    Computes the directed Hausdorff distance of Q_1 and Q_2 (AH_polytopes)
    ***************************************************************************
    The optimization problem is:
                        Minimize    epsilon  
                        such that   Q1 \subset Q2+epsilon(Ball)
                        
    It is zero if and only if Q1 subset Q2. The method is based on 
                
                    Sadraddini&Tedrake, 2019, CDC (available on ArXiv)
                    
    We solve the following problem:
        D*ball+Q1 subset Q2
    We solve the following linear program:
    ..math::
        \min     D
        s.t.    Lambda_1 H_1=H_2 Gamma_1
                Lambda_2 H_1=H_ball Gamma_2
                Lambda_1 h_1<=h_2 + H_2 beta_1
                Lambda_2 h_2<=D h_ball + H_ball beta_2
                x_2 - X_2 beta_1 - beta_2 = x_1
                X_2 Gamma_1 + Gamma_2 = X_1
    ***************************************************************************
    """
    Q1,Q2=pp.to_AH_polytope(Q1),pp.to_AH_polytope(Q2)
    n=Q1.t.shape[0]
    if ball=="infinty_norm":
        HB=np.vstack((np.eye(n),-np.eye(n)))
        hB=np.vstack((np.ones((n,1)),np.ones((n,1))))
    elif ball=="l1":
        HB,hb=make_ball(ball)
    prog=MP.MathematicalProgram()
    # Variables
    D=prog.NewContinuousVariables(1,1,"D")
    Lambda_1=prog.NewContinuousVariables(Q2.P.H.shape[0],Q1.P.H.shape[0],"Lambda_1")
    Lambda_2=prog.NewContinuousVariables(HB.shape[0],Q1.P.H.shape[0],"Lambda2")
    Gamma_1=prog.NewContinuousVariables(Q2.P.H.shape[1],Q1.P.H.shape[1],"Gamma1")
    Gamma_2=prog.NewContinuousVariables(HB.shape[1],Q1.P.H.shape[1],"Gamma1")
    beta_1=prog.NewContinuousVariables(Q2.P.H.shape[1],1,"beta1")
    beta_2=prog.NewContinuousVariables(HB.shape[1],1,"beta1")
    # Constraints
    # Lambda_1 and Lambda_2 positive
    prog.AddBoundingBoxConstraint(0,np.inf,Lambda_1)
    prog.AddBoundingBoxConstraint(0,np.inf,Lambda_2)
    # Lambda_1 H_1
    Lambda_H_Gamma(prog,Lambda_1,Q1.P.H,Q2.P.H,Gamma_1)
    # Lambda_2 H_1
    Lambda_H_Gamma(prog,Lambda_2,Q1.P.H,HB,Gamma_2)
    # Lambda_1 h_1
    Lambda_h_Inequality(prog,Lambda_1,beta_1,Q2.P.H,Q1.P.h,Q2.P.h)
    # Lambda_2 h_1
    Lambda_h_Inequality_D(prog,Lambda_2,beta_2,HB,Q1.P.h,hB,D)
    # X2 beta_1   
    prog.AddLinearEqualityConstraint(-np.hstack((Q2.T,np.eye(n))),Q1.t-Q2.t,np.vstack((beta_1,beta_2)))
    # X2 Gamma_1
    Aeq=np.hstack((Q2.T,np.eye(Q2.T.shape[0])))
    for i in range(Gamma_1.shape[1]):
        beq=Q1.T[:,i]
        var=np.hstack((Gamma_1[:,i],Gamma_2[:,i]))
        prog.AddLinearEqualityConstraint(Aeq,beq,var)
    # Cost
    # Optimize
    if solver=="gurobi":
            prog.AddLinearCost(D[0,0])
            result=gurobi_solver.Solve(prog,None,None)
    elif solver=="osqp":
        prog.AddQuadraticCost(D[0,0]*D[0,0])
        result=OSQP_solver.Solve(prog,None,None)
    else:
        result=MP.Solve(prog)
    if result.is_success():
        return np.asscalar(result.GetSolution(D))

def Hausdorff_distance(Q1,Q2,ball="infinty_norm",solver="gurobi"):
    return max(directed_Hausdorff_distance(Q1,Q2,ball,solver),directed_Hausdorff_distance(Q2,Q1,ball,solver))
    
def distance_polytopes(Q1,Q2,ball="infinity",solver="gurobi"):
    Q1,Q2=pp.to_AH_polytope(Q1),pp.to_AH_polytope(Q2)
    n=Q1.n
    prog=MP.MathematicalProgram()
    zeta1=prog.NewContinuousVariables(Q1.P.H.shape[1],1,"zeta1")
    zeta2=prog.NewContinuousVariables(Q2.P.H.shape[1],1,"zeta2")
    delta=prog.NewContinuousVariables(n,1,"delta")
    prog.AddLinearConstraint(A=Q1.P.H,ub=Q1.P.h,lb=-np.inf*np.ones((Q1.P.h.shape[0],1)),vars=zeta1)
    prog.AddLinearConstraint(A=Q2.P.H,ub=Q2.P.h,lb=-np.inf*np.ones((Q2.P.h.shape[0],1)),vars=zeta2)
    prog.AddLinearEqualityConstraint( np.hstack((Q1.T,-Q2.T,np.eye(n))),Q2.t-Q1.t,np.vstack((zeta1,zeta2,delta)) )
    if ball=="infinity":
        delta_abs=prog.NewContinuousVariables(1,1,"delta_abs")
        prog.AddBoundingBoxConstraint(0,np.inf,delta_abs)
        prog.AddLinearConstraint(np.greater_equal( np.dot(np.ones((n,1)),delta_abs),delta,dtype='object' ))
        prog.AddLinearConstraint(np.greater_equal( np.dot(np.ones((n,1)),delta_abs),-delta,dtype='object' ))
        cost=delta_abs
    elif ball=="l1":
        delta_abs=prog.NewContinuousVariables(n,1,"delta_abs")
        prog.AddBoundingBoxConstraint(0,np.inf,delta_abs)
        prog.AddLinearConstraint(np.greater_equal( delta_abs,delta,dtype='object' ))
        prog.AddLinearConstraint(np.greater_equal( delta_abs,-delta,dtype='object' ))
        cost=np.dot(np.ones((1,n)),delta_abs)
    else:
        raise NotImplementedError
    if solver=="gurobi":
        prog.AddLinearCost(cost[0,0])
        result=gurobi_solver.Solve(prog,None,None)
    elif solver=="osqp":
        prog.AddQuadraticCost(cost[0,0]*cost[0,0])
        result=OSQP_solver.Solve(prog,None,None)
    else:
        prog.AddLinearCost(cost[0,0])
        result=MP.Solve(prog)
    if result.is_success():
        return np.sum(result.GetSolution(delta_abs)),\
            np.dot(Q1.T,result.GetSolution(zeta1).reshape(zeta1.shape[0],1))+Q1.t,\
            np.dot(Q2.T,result.GetSolution(zeta2).reshape(zeta2.shape[0],1))+Q2.t

def _setup_program_distance_point(P,ball="infinity",solver="Gurobi"):
    """
    Initilize the mathematial program
    Choice of balls:
        infinity: L-infinity norm
        l1: l1 norm (Manhattan Distance)
        l2: l2 norm (Euclidean Distance)
    """
    if P.distance_program is None:
        prog=MP.MathematicalProgram()
        Q=pp.to_AH_polytope(P)
        n=Q.n
        x=np.zeros((n,1))
        P.zeta=prog.NewContinuousVariables(Q.P.H.shape[1],1,"zeta")
        delta=prog.NewContinuousVariables(n,1,"delta")
        prog.AddLinearConstraint(A=Q.P.H,ub=Q.P.h,lb=-np.inf*np.ones((Q.P.h.shape[0],1)),vars=P.zeta)
        P.distance_constraint=prog.AddLinearEqualityConstraint( np.hstack((Q.T,-np.eye(n))),x-Q.t,np.vstack((P.zeta,delta)) )
        if ball=="infinity":
            delta_abs=prog.NewContinuousVariables(1,1,"delta_abs")
            prog.AddBoundingBoxConstraint(0,np.inf,delta_abs)
            prog.AddLinearConstraint(np.greater_equal( np.dot(np.ones((n,1)),delta_abs),delta,dtype='object' ))
            prog.AddLinearConstraint(np.greater_equal( np.dot(np.ones((n,1)),delta_abs),-delta,dtype='object' ))
            prog.AddLinearCost(delta_abs[0,0])
        elif ball=="l1":
            delta_abs=prog.NewContinuousVariables(n,1,"delta_abs")
            prog.AddBoundingBoxConstraint(0,np.inf,delta_abs)
            prog.AddLinearConstraint(np.greater_equal( delta_abs,delta,dtype='object' ))
            prog.AddLinearConstraint(np.greater_equal( delta_abs,-delta,dtype='object' ))
            cost=np.dot(np.ones((1,n)),delta_abs)
            prog.AddLinearCost(cost[0,0])
        elif ball=="l2":
            prog.AddQuadraticCost(np.eye(n),np.zeros(n),delta)
        else:
            print(("Not a valid choice of norm",str(ball)))
            raise NotImplementedError
        P.distance_program=prog
        return 
    else:
        return
            
        
def distance_point_polytope(P, x, ball="infinity", solver="Gurobi"):
    """
    Computes the distance of point x from AH-polytope Q 
    """
    x_vector = np.atleast_2d(x) #in case x is not n*1 vector
    P = pp.to_AH_polytope(P)
    _setup_program_distance_point(P,ball,solver)
    prog=P.distance_program
    Q=pp.to_AH_polytope(P)
    a=P.distance_constraint.evaluator()
    x_vector=x_vector.reshape(max(x_vector.shape),1)
    a.UpdateCoefficients(np.hstack((Q.T,-np.eye(Q.n))), x_vector - Q.t)
    if solver=="Gurobi":
        result=gurobi_solver.Solve(prog,None,None)
    elif solver=="osqp":
        result=OSQP_solver.Solve(prog,None,None)
    else:
        result=MP.Solve(prog)
    if result.is_success():
        zeta_num=result.GetSolution(P.zeta).reshape(P.zeta.shape[0],1)
        x_nearest=np.dot(Q.T,zeta_num)+Q.t
        delta=(x_vector - x_nearest).reshape(Q.n)
        if ball=="infinity":
            d=np.linalg.norm(delta,ord=np.inf)
        elif ball=="l1":
            d=np.linalg.norm(delta,ord=1)
        elif ball=="l2":
            d=np.linalg.norm(delta,ord=2)  
        return d,x_nearest
    
def bounding_box(Q,solver="Gurobi"):
    r"""
    Computes the bounding box of a polytope by solving :math:`2n` linear programs. 
    Each linear program is in the form:
        
    .. math:: 
        \begin{array}{lll}
        l_i= & \min & e_i^T x \\
        & \text{subject to} & x \in \mathbb{P}
        \end{array}
        
    and
    
    .. math:: 
        \begin{array}{lll}
        u_i= & \max & e_i^T x \\
        & \text{subject to} & x \in \mathbb{P}
        \end{array}  
        
    where :math:`l,u` define the lower and upper corners of the bounding box.
    """
    Q=pp.to_AH_polytope(Q)
    prog=MP.MathematicalProgram()
    zeta=prog.NewContinuousVariables(Q.P.H.shape[1],1,"zeta")
    x=prog.NewContinuousVariables(Q.n,1,"x")
    prog.AddLinearConstraint(A=Q.P.H,ub=Q.P.h,lb=-np.inf*np.ones((Q.P.h.shape[0],1)),vars=zeta)
    prog.AddLinearEqualityConstraint(np.hstack((-Q.T,np.eye(Q.n))),Q.t,np.vstack((zeta,x)))
    lower_corner=np.zeros((Q.n,1))
    upper_corner=np.zeros((Q.n,1))
    c=prog.AddLinearCost(np.ones(Q.n),0,x)
    if solver=="Gurobi":
        solver=gurobi_solver
    else:
        raise NotImplementedError
    a=np.zeros((Q.n,1))
    # Lower Corners
    for i in range(Q.n):
        e=c.evaluator()
        a[i,0]=1
        e.UpdateCoefficients(a.reshape(Q.n))
        result=solver.Solve(prog,None,None)
        assert result.is_success()
        lower_corner[i,0]=result.GetSolution(x)[i]
        a[i,0]=0
    # Upper Corners
    for i in range(Q.n):
        e=c.evaluator()
        a[i,0]=-1
        e.UpdateCoefficients(a.reshape(Q.n))
        result=solver.Solve(prog,None,None)
        assert result.is_success()
        upper_corner[i,0]=result.GetSolution(x)[i]
        a[i,0]=0
    return pp.hyperbox(corners=(lower_corner,upper_corner))
        
        
def directed_Hausdorff_hyperbox(b1,b2):
    """
    The directed Hausdorff hyperbox 
    min epsilon such that b1 \in b2+epsilon
    """       
    return max(0,np.max(np.hstack((b1.u-b2.u,b2.l-b1.l))))           
    
def distance_hyperbox(b1,b2):
    """
    The distance between boxes
    """
    return max(0,np.max(np.hstack((b1.l-b2.u,b2.l-b1.u))))      
    

def make_ball(n,norm):
    # TO COMPLETE
    if norm=="l1":
        pass
    elif norm=="infinity":
        pass
    return 

def get_nonzero_cost_vectors(cost):
     cost[cost == 0] = 1e-3

def AH_polytope_vertices(P,N=200,epsilon=0.001,solver="Gurobi"):
    """
    Returns N*2 matrix of vertices
    """
    try:
        P.vertices_2D
        if type(P.vertices_2D) == type(None):
            raise Exception
    except:
        Q=pp.to_AH_polytope(P)
        v=np.empty((N,2))
        prog=MP.MathematicalProgram()
        zeta=prog.NewContinuousVariables(Q.P.H.shape[1],1,"zeta")
        prog.AddLinearConstraint(A=Q.P.H,ub=Q.P.h,lb=-np.inf*np.ones((Q.P.h.shape[0],1)),vars=zeta)
        theta=1
        c=np.array([np.cos(theta),np.sin(theta)]).reshape(2,1)
        c_T=np.dot(c.T,Q.T)
        get_nonzero_cost_vectors(c_T)
        # get_nonzero_cost_vectors(c_T)
        a=prog.AddLinearCost(np.dot(c_T,zeta)[0,0])
        if solver=="Gurobi":
            solver=gurobi_solver
        else:
            raise NotImplementedError
        for i in range(N):
            theta=i*N/2/np.pi+0.01
            c=np.array([np.cos(theta),np.sin(theta)]).reshape(2,1)
            c_T=np.dot(c.T,Q.T)
            e=a.evaluator()
            cost = c_T.reshape(Q.P.H.shape[1])
            get_nonzero_cost_vectors(cost)
            e.UpdateCoefficients(cost)
            result=solver.Solve(prog,None,None)
            assert result.is_success()
            zeta_n=result.GetSolution(zeta).reshape(zeta.shape)
            v[i,:]=(np.dot(Q.T,zeta_n)+Q.t).reshape(2)
        w=np.empty((4*N,2))
        for i in range(N):
            w[4*i,:]=v[i,:]+np.array([epsilon,epsilon])
            w[4*i+1,:]=v[i,:]+np.array([-epsilon,epsilon])
            w[4*i+2,:]=v[i,:]+np.array([-epsilon,-epsilon])
            w[4*i+3,:]=v[i,:]+np.array([epsilon,-epsilon])
        P.vertices_2D=v,w
        return v,w
    else:
        return P.vertices_2D
    
def convex_hull_of_point_and_polytope(x, Q):
    r"""
    Inputs:
        x: numpy n*1 array
        Q: AH-polytope in R^n
    Returns:
        AH-polytope representing convexhull(x,Q)
    
    .. math::
        \text{conv}(x,Q):=\{y | y= \lambda q + (1-\lambda) x, q \in Q\}.
    """
    Q=pp.to_AH_polytope(Q)
    q=Q.P.H.shape[1]
    new_T=np.hstack((Q.T,Q.t-x))
    new_t=x
    new_H_1=np.hstack((Q.P.H,-Q.P.h))
    new_H_2=np.zeros((2,q+1))
    new_H_2[0,q],new_H_2[1,q]=1,-1
    new_H=np.vstack((new_H_1,new_H_2))
    new_h=np.zeros((Q.P.h.shape[0]+2,1))
    new_h[Q.P.h.shape[0],0],new_h[Q.P.h.shape[0]+1,0]=1,0
    new_P=pp.H_polytope(new_H,new_h)
    return pp.AH_polytope(new_T,new_t,new_P)
    

def minkowski_sum(P1,P2):
    r"""
    Inputs: 
        P1, P2: AH_polytopes
    Returns:
        returns the Mkinkowski sum :math:`P_1 \oplus P_2` as an AH-polytope.
        
    **Background**: The Minkowski sum of two sets is defined as:
        
    .. math::
        A \oplus B = \{ a + b \big | a \in A, b \in B\}.
    """
    Q1,Q2=pp.to_AH_polytope(P1),pp.to_AH_polytope(P2)
    T=np.hstack((Q1.T,Q2.T))
    t=Q1.t+Q2.t
    H=spa.block_diag(*[Q1.P.H,Q2.P.H])
    h=np.vstack((Q1.P.h,Q2.P.h))
    new_P=pp.H_polytope(H,h)
    return pp.AH_polytope(t=t,T=T,P=new_P)  
    
    
def intersection_old(P1,P2):
    """
    Inputs: 
        P1, P2: AH_polytopes :math:`\mathbb{P}_1,\mathbb{P}_2`. Converted to AH-polytopes
    Output:
        returns :math:`\mathbb{P}_1 \cap \mathbb{P}_2` as an AH-polytope
    """
    Q1,Q2=pp.to_AH_polytope(P1),pp.to_AH_polytope(P2)
    T=np.hstack((Q1.T,Q2.T*0))
    t=Q1.t
    H_1=spa.block_diag(*[Q1.P.H,Q2.P.H])
    H_2=np.hstack((Q1.T,-Q2.T))
    H=np.vstack((H_1,H_2,-H_2))
    h=np.vstack((Q1.P.h,Q2.P.h,Q2.t-Q1.t,Q1.t-Q2.t))
    new_P=pp.H_polytope(H,h)
    return pp.AH_polytope(T=T,t=t,P=new_P)

def convex_hull(P1,P2):
    """
    Inputs:
        P1, P2: AH_polytopes
    Output:
        returns :math:`\text{ConvexHull}(\mathbb{P}_1,\mathbb{P}_2)` as an AH-polytope
    """
    Q1,Q2=pp.to_AH_polytope(P1),pp.to_AH_polytope(P2)
    T=np.hstack((Q1.T,Q2.T,Q1.t-Q2.t ))
    H_1 = np.hstack((Q1.P.H,np.zeros((Q1.P.H.shape[0],Q2.P.n)), -Q1.P.h ))
    H_2 = np.hstack((np.zeros((Q2.P.H.shape[0],Q1.P.n)), Q2.P.H, Q2.P.h ))
    H_3 = np.zeros(( 2, Q1.P.n + Q2.P.n +1 ))
    H_3[:,-1:]=np.array([1,-1]).reshape(2,1)
    H=np.vstack((H_1,H_2,H_3))
    h=np.vstack((Q1.P.h*0,Q2.P.h,1,0))
    new_P=pp.H_polytope(H,h)
    return pp.AH_polytope(T=T,t=Q2.t,P=new_P)

def intersection(P1,P2):
    """
    Inputs:
        P1, P2: polytopic objects
    Output:
        returns :math:`\mathbb{P}_1 \cap \mathbb{P}_2` as an AH-polytope
        
        If both objects are H-polytopes, return H-polytope
    """
    if P1.type=="H_polytope" and P2.type=="H_polytope":
        H=np.vstack((P1.H,P2.H))
        h=np.vstack((P1.h,P2.h))
        return pp.H_polytope(H,h)
    else:
        X,Y=pp.to_AH_polytope(P1),pp.to_AH_polytope(P2)
        T=np.hstack(( X.T, np.zeros((X.T.shape[0]  ,  Y.T.shape[1] ))  ))
        H_1 = np.hstack((X.P.H,   np.zeros((X.P.H.shape[0],Y.P.H.shape[1])) ))
        Ty_inv=np.linalg.pinv(Y.T)
        H_2 = np.hstack(( np.linalg.multi_dot([Y.P.H,Ty_inv,X.T]),\
                         np.dot(Y.P.H,np.eye(Y.T.shape[1])-np.dot(Ty_inv,Y.T))  ))
        H=np.vstack((H_1,H_2))
        h=np.vstack((X.P.h,Y.P.h-np.linalg.multi_dot([Y.P.H,Ty_inv,X.t-Y.t])))
        new_P=pp.H_polytope(H,h)
        return pp.AH_polytope(T=T,t=X.t,P=new_P)

    
"""
Pydrake Mathematical Program Helper: Matrix based Constraints
"""
def AddMatrixInequalityConstraint_classical(mathematical_program,A,X,B):
    raise NotImplementedError    
    
def Lambda_h_Inequality(mathematical_program,Lambda,beta,H,h_1,h_2):
    """
    Adds Lambda H-1 \le h_2 + H beta to the Mathematical Program
    """
    M1=np.kron(np.eye(h_2.shape[0]),h_1.T)
    M=np.hstack((M1,-H))
    v1=Lambda.reshape((Lambda.shape[0]*Lambda.shape[1],1))
    v=np.vstack((v1,beta))
    mathematical_program.AddLinearConstraint(A=M,ub=h_2,lb=-np.inf*np.ones(h_2.shape),vars=v)
    
def Lambda_h_Inequality_D(mathematical_program,Lambda,beta,H,h_1,h_2,D):
    """
    Adds Lambda H-1 \le h_2 D + H beta to the Mathematical Program
    """
    M1=np.kron(np.eye(h_2.shape[0]),h_1.T)
    M=np.hstack((M1,-H,-h_2))
    v1=Lambda.reshape((Lambda.shape[0]*Lambda.shape[1],1))
    v=np.vstack((v1,beta,D))
    mathematical_program.AddLinearConstraint(A=M,ub=np.zeros(h_2.shape),lb=-np.inf*np.ones(h_2.shape),vars=v)
    
def positive_matrix(mathematical_program,Lambda):
    """
    All elements are non-negative 
    """
#    q=Lambda.shape[0]
    mathematical_program.AddBoundingBoxConstraint(0,np.inf,Lambda)
#    [mathematical_program.AddLinearConstraint(A=np.eye(q),vars=Lambda[:,i],
#                                              ub=np.inf*np.ones((q,1)),lb=np.zeros((q,1)))
#                                                for i in range(Lambda.shape[1])]
#    q=Lambda.shape[0]*Lambda.shape[1]
#    mathematical_program.AddLinearConstraint(A=np.eye(q),vars=Lambda.reshape(q),ub=np.inf*np.ones((q)),lb=np.zeros((q)))
        
def Lambda_H_Gamma(mathematical_program,Lambda,H_1,H_2,Gamma):
    """
    Lambda H_1 = H_2 Gamma
    """
#    v_1=Lambda.reshape((Lambda.shape[0]*Lambda.shape[1],1))
#    for j in range(Gamma.shape[1]):
#        M1=np.kron(np.eye(H_2.shape[0]),H_1[:,j].reshape(1,H_1.shape[0]))
#        M2=-H_2
#        M=np.hstack((M1,M2))
#        v=np.vstack((v_1,Gamma[:,j].reshape(Gamma.shape[0],1)))
#        mathematical_program.AddLinearEqualityConstraint(M,np.zeros((M.shape[0],1)),v)
    for i in range(Lambda.shape[0]):
        for j in range(Gamma.shape[1]):
            M=np.hstack((H_1[:,j],-H_2[i,:]))
            v=np.hstack((Lambda[i,:],Gamma[:,j]))
            mathematical_program.AddLinearEqualityConstraint(M.reshape(1,M.shape[0]),np.zeros(1),v)




def sorting_generator(G,desired_numberofcolumns):
    """
    The goal is deviding the generator into to parts.
    One part that is used for zonotope order reduction methods.
    And the other part which is used to enforce the reduced zonotope to have the desire order.
    """
    dimension = len(G)

    if dimension == desired_numberofcolumns:
        return G , None
    indexlist = np.argsort(np.linalg.norm( G ,axis=0 ,ord = None))
    sortedG = (G)[:,indexlist]
    G_untouched = sortedG[: , - (desired_numberofcolumns - dimension ): ]
    G_reduced = sortedG[: , : -(desired_numberofcolumns - dimension )  ]
    
    return G_reduced , G_untouched


def boxing_order_reduction(zonotope,desired_order=1):
    """
    boxing method for zonotope order reduction
    inputs: input zonotope , order of the output zonotope
    output: zonotope

    Based on Kopetzki, Anna-Kathrin, Bastian Schürmann, and Matthias Althoff.
    "Methods for order reduction of zonotopes."
    2017 IEEE 56th Annual Conference on Decision and Control (CDC). IEEE, 2017.
    """
    assert(type(zonotope) == pp.zonotope),"TypeError: The first argument need to be from \"zonotope\" class in pypolycontain package "
    assert(type(desired_order) == int or type(desired_order) == float) , "TypeError: The second argument need to be a number greater than 1. \n \
                                                                            It is the order of the reduced zonotpe which equal to the number of \
                                                                                columns over the space dimension."
    assert(desired_order >= 1), "desired order of the outcome zonotope needs to be greater or equal to 1"
    
    x = np.array(zonotope.x)
    G = np.array(zonotope.G)
    
    dimension , numberofcolumns =  G.shape
    desired_numberofcolumns = round(desired_order * dimension)
    
    if numberofcolumns <= desired_numberofcolumns:
        return zonotope
    
    elif dimension == desired_numberofcolumns:
        G_box = np.diag(np.sum(abs( G ) ,axis=1 ))
        return pp.zonotope( G_box , x)  
    
    else:
        G_reduced , G_untouched = sorting_generator( G , desired_numberofcolumns )
        G_box = np.concatenate(  ( np.diag(np.sum(abs( G_reduced ) ,axis=1 )) , G_untouched   ), axis=1  )
        return pp.zonotope( G_box , x)  


def pca_order_reduction(zonotope,desired_order=1):

    """
    PCA method for zonotope order reduction
    inputs: input zonotope , order of the output zonotope
    output: zonotope

    Based on Kopetzki, Anna-Kathrin, Bastian Schürmann, and Matthias Althoff.
    "Methods for order reduction of zonotopes."
    2017 IEEE 56th Annual Conference on Decision and Control (CDC). IEEE, 2017.
    """
    assert(type(zonotope) == pp.zonotope),"TypeError: The first argument need to be from \"zonotope\" class in pypolycontain package "
    assert(type(desired_order) == int or type(desired_order) == float) , "TypeError: The second argument need to be a number greater than 1. \n \
                                                                            It is the order of the reduced zonotpe which equal to the number of \
                                                                                columns over the space dimension."
    assert(desired_order >= 1), "desired order of the outcome zonotope needs to be greater or equal to 1"
    
    x = np.array(zonotope.x)
    G = np.array(zonotope.G)
    
    dimension , numberofcolumns =  G.shape
    desired_numberofcolumns = round(desired_order * dimension)

    if numberofcolumns <= desired_numberofcolumns:
        return zonotope

    G_reduced , G_untouched = sorting_generator( G , desired_numberofcolumns )
    X = np.concatenate(  (G_reduced ,  - G_reduced), axis=1 ).T
    covariance = np.dot(X.T , X)
    U,_,_ =np.linalg.svd( covariance )
    interval_hull = boxing_order_reduction( pp.zonotope( np.dot(U.T , G_reduced) , x ) ).G 
    G_pca = np.dot(U , interval_hull)
    
    if type(G_untouched)== np.ndarray:
        return pp.zonotope( np.concatenate((  G_pca , G_untouched   ), axis=1  ) , x) 
    elif G_untouched == None:
        return pp.zonotope(G_pca , x)


def decompose(zonotope,dimensions):
    """
    @author: kasra
    Decompising a given set into bunch of fewer dimensional sets such that 
    the Cartesian product of those sets is a subset of the given set.
    """

    assert(sum(dimensions) == len(zonotope.G)), "ValueError: sum of the given dimensions has \
                                                to be equal to the dimension of the input set"
    #number_of_sets = len(dimensions)

    prog=MP.MathematicalProgram()

    #defining varibales
    x_i = [prog.NewContinuousVariables(i) for i in dimensions]
    G_i = [prog.NewContinuousVariables(i,i) for i in dimensions]                #non-symmetric G_i
    #G_i = [prog.NewSymmetricContinuousVariables(i) for i in dimensions ]               #symmentric G_i

    #inbody_zonotope
    inbody_x = np.concatenate(x_i)
    inbody_G = block_diag(*G_i)
    inbody_zonotope = pp.zonotope(inbody_G , inbody_x)

    #Defining Constraints
    prog.AddPositiveSemidefiniteConstraint(inbody_G)
    pp.subset(prog, inbody_zonotope , zonotope)
    # ASSUMPTION for PSD of inbody_G
    # ASSUMPTION for SYMMENTRICITY of G_i

    #Defining the cost function
    prog.AddMaximizeLogDeterminantSymmetricMatrixCost( inbody_G )

    #Solving
    result = MP.Solve(prog)
    
    print(f"Is optimization successful? {result.is_success()}")
    print(f"Solution to x_i: {result.GetSolution(x_i[0])}")
    print(f"Solution to G_i: {result.GetSolution(G_i[0])}")
    print(f"optimal cost: {result.get_optimal_cost()}")
    print('solver is: ', result.get_solver_id().name())

    x_i_result = [result.GetSolution(x_i[i]) for i in range(len(dimensions))]
    G_i_result = [result.GetSolution(G_i[i]) for i in range(len(dimensions))]
    list_zon = [pp.zonotope(G = G_i_result[i],x = x_i_result[i]) for i in range(len(dimensions))]

    return list_zon


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 23 23:35:22 2020

@author: sadra
"""

import numpy as np
from itertools import combinations
import pypolycontain as pp

G=np.random.random((2,8))-0.5
G=np.array([[1,0,0],[0,1,0]])
Z=pp.zonotope(G,color='red')


S=combinations(range(G.shape[1]),G.shape[0])
V=0
for s in S:
#    print(s)
    Gs=np.hstack([G[:,i:i+1] for i in s])
    print(Gs)
    V+=abs(np.linalg.det(Gs))
V*=2**G.shape[0]
print(V)

pp.visualize([Z,pp.zonotope(np.eye(2)*0.5)],title=r'volume=%0.002f'%V)

S=combinations(range(G.shape[1]),G.shape[0])
V_dot=np.zeros(G.shape)
for s in S:
    Gs=np.hstack([G[:,i:i+1] for i in s])
    D=np.linalg.det(Gs)
    if D!=0:
        adj_Gs=D*np.linalg.inv(Gs)
    else:
        print("Warning: we have a singular matrix")
        e=np.eye(G.shape[0])*10**(-5)
        adj_Gs=np.linalg.det(Gs+e)*np.linalg.inv(Gs+e)
    X=adj_Gs.T*np.sign(np.linalg.det(Gs+e))
    for i in range(len(s)):
        V_dot[:,s[i]]+=X[:,i]
print(V_dot)


import numpy as np
import pypolycontain as pp
#import warnings

from matplotlib.patches import Polygon
import matplotlib.pyplot as plt
from matplotlib.collections import PatchCollection

# Pypolycontain
#try:
#    from pypolycontain.conversions import to_V    
#except:
#    pass
#    warnings.warn("You don't have pypolycontain properly installed. Can not import conversions")
    


def visualize(list_of_objects,fig=None,ax=None,a=0.5,alpha=0.8,tuple_of_projection_dimensions=[0,1],\
              title=r'pypolycontain visualization',\
              show_vertices=False,FontSize=15,equal_axis=False,grid=True,\
              N_points=1000):
    r"""
    Visualization.
    
    inputs: 
        * list_of_objects:
        * fig:
        * tuple_of_projection_dimensions: 
        * 
    """
    if type(ax)==type(None):
        fig,ax=plt.subplots()
        fig.figsize=(15,10)
    p_list,x_all=[],np.empty((0,2))  
    for p in list_of_objects:
        x=pp.to_V(p,N=N_points)
        mypolygon=Polygon(x)
        p_list.append(mypolygon) 
        x_all=np.vstack((x_all,x))
        if show_vertices:
            ax.plot(x[:,0],x[:,1],'*',color=p.color)
    p_patch = PatchCollection(p_list,color=[p.color for p in list_of_objects], alpha=alpha)
    ax.add_collection(p_patch)
    ax.set_xlim([np.min(x_all[:,0])-a,a+np.max(x_all[:,0])])
    ax.set_ylim([np.min(x_all[:,1])-a,a+np.max(x_all[:,1])])
    if grid:
        ax.grid(color=(0,0,0), linestyle='--', linewidth=0.3)
    ax.set_title(title,FontSize=FontSize)
    ax.set_xlabel(r"$x_{%d}$"%(tuple_of_projection_dimensions[0]+1),FontSize=FontSize)
    ax.set_ylabel(r"$x_{%d}$"%(tuple_of_projection_dimensions[1]+1),FontSize=FontSize)
    if equal_axis:
        ax.axis('equal')


import warnings
import numpy as np
# Scipy
try:
    from scipy.spatial import ConvexHull
except:
    warnings.warn("You don't have scipy package installed. You may get error while using some feautures.")
    
# Pypolycontain
try:
    import pypolycontain as pp
except:
    warnings.warn("You don't have pypolycontain properly installed. Can not import objects")
    
# pycdd
try:
    from cdd import Polyhedron,Matrix,RepType
except:
    warnings.warn("WARNING: You don't have CDD package installed. Unable to visualize polytopes. You may still visualize zonotopes.")

# Pydrake
try:
    import pydrake.solvers.mathematicalprogram as MP
    import pydrake.solvers.gurobi as Gurobi_drake
    import pydrake.solvers.osqp as OSQP_drake
    # use Gurobi solver
    global gurobi_solver,OSQP_solver, license
    gurobi_solver=Gurobi_drake.GurobiSolver()
    license = gurobi_solver.AcquireLicense()
    OSQP_solver=OSQP_drake.OsqpSolver()
except:
    warnings.warn("You don't have pydrake installed properly. Methods that rely on optimization may fail.")

try:
    import pydrake.solvers.scs as scs_drake
    scs_solver=scs_drake.ScsSolver()
except:
    warnings.warn("You don't have pydrake with SCS solver.")
    

def to_AH_polytope(P):
    """
    Converts the polytopic object P into an AH-polytope. If applied on 
    a AH-polytope, a deepcopy is returned
    """
    if type(P).__name__=="AH_polytope":
        return pp.AH_polytope(T=P.T,t=P.t,P=pp.H_polytope(P.P.H,P.P.h))
    elif type(P).__name__=="H_polytope":
        n=P.H.shape[1]
        return pp.AH_polytope(T=np.eye(n),t=np.zeros((n,1)),P=pp.H_polytope(P.H,P.h))
    elif type(P).__name__=="zonotope":
        q=P.G.shape[1]
        return pp.AH_polytope(T=P.G,t=P.x.reshape(-1,1),P=pp.unitbox(N=q).H_polytope,color=P.color)
    elif type(P).__name__=="V_polytope":
        V=P.list_of_vertices
        N=len(V)
        T=np.hstack([V[i]-V[-1] for i in range(N-1)])
        t=V[-1]
        H=np.vstack((-np.eye(N-1),np.ones((1,N-1))))
        h=np.zeros((N,1))
        h[N-1,0]=1
        P=pp.H_polytope(H,h)
        return pp.AH_polytope(t,T,P)
    else:
        raise ValueError("object type not within my polytopic library:",P.type)

def H_to_V(P):
    r"""
    Returns the vertices of an H_polytope.
    
    Inputs:
        * P: H_polytope in :math:`\mathbb{R}^n`
    Output:
        * V: list of vertices. Each vertex is *numpy.ndarray[float[n,1]]*
    
    **Method**:
        The method is based on double description method, and is using pycddlib.
    .. warning::
        This method can be very slow or numerically unstable for polytopes in high dimensions and/or large number of hyperplanes
    """
    if type(P).__name__=="H_polytope":
        p_mat=Matrix(np.hstack((P.h,-P.H)))
        p_mat.rep_type = RepType.INEQUALITY
        poly=Polyhedron(p_mat)
        y=np.array(poly.get_generators())
        x=y[:,1:]#
        x=x[ConvexHull(x).vertices,:]
        return x
    else:
        raise ValueError(str(type(P).__name__)+" is not an H-polytope")
        

def AH_to_V(P,N=360,epsilon=1e-3,solver="Gurobi"):
    """
    Returns the V-polytope form of a 2D pp.AH_polytope.
    The method is based on ray shooting. 
    
    Inputs:
        * P: AH-polytope
        * N ``defualt=360``: number of rays
        * solver: ``default=Gurobi``. The linear-programming optimization solver.
    Returns:
        * V: matrix 
        
    .. note::
        This method only works for 2D AH-polytopes and its for visualization. 
        For generic use, first use H-V on :math:`\\mathbb{P}` and then apply affine transformation.
        Note that H-V uses elimination-based vertex enumeration method that is not scalable.
    """
    Q=pp.to_AH_polytope(P)
    if Q.n!=2:
        raise ValueError("Sorry, but I can only do AH to V operation in 2D using ray shooting")
    v=np.empty((N,2))
    prog=MP.MathematicalProgram()
    zeta=prog.NewContinuousVariables(Q.P.H.shape[1],1,"zeta")
    prog.AddLinearConstraint(A=Q.P.H,ub=Q.P.h,lb=-np.inf*np.ones((Q.P.h.shape[0],1)),vars=zeta)
    theta=1
    c=np.array([np.cos(theta),np.sin(theta)]).reshape(2,1)
    c_T=np.dot(c.T,Q.T)
    c_T[c_T == 0] = 1e-3
    # get_nonzero_cost_vectors(c_T)
    a=prog.AddLinearCost(c_T.reshape(Q.P.n),np.zeros(1),zeta)
    if solver=="Gurobi":
        solver=gurobi_solver
    elif solver=="SCS":
        solver=scs_solver
    else:
        raise NotImplementedError
    for i in range(N):
        theta=i*N/2/np.pi+0.01
        c=np.array([np.cos(theta),np.sin(theta)]).reshape(2,1)
        c_T=np.dot(c.T,Q.T)
        e=a.evaluator()
        cost = c_T.reshape(Q.P.H.shape[1])
#        cost[cost == 0] = 1e-3
        e.UpdateCoefficients(cost)
        result=solver.Solve(prog,None,None)
        assert result.is_success()
        zeta_n=result.GetSolution(zeta).reshape(zeta.shape)
        v[i,:]=(np.dot(Q.T,zeta_n)+Q.t).reshape(2)
    try:
        v=v[ConvexHull(v).vertices,:]
        return v
    except: # convexhull very small. Add some epsilon
        w=np.empty((4*N,2))
        for i in range(N):
            w[4*i,:]=v[i,:]+np.array([epsilon,epsilon])
            w[4*i+1,:]=v[i,:]+np.array([-epsilon,epsilon])
            w[4*i+2,:]=v[i,:]+np.array([-epsilon,-epsilon])
            w[4*i+3,:]=v[i,:]+np.array([epsilon,-epsilon])
        w=w[ConvexHull(w).vertices,:]
        return w
    
def zonotope_to_V(Z):
    """
    Finds the vertices of a zonotope
    """
    q=Z.G.shape[1]
    if q<13:
        v=Z.x.T+np.dot(Z.G,vcube(q).T).T
        return v[ConvexHull(v).vertices,:]
    else:
        warnings.warn('Zonotope Vertex Enumeration: \
                      The number of generators %d is very large. \
                      Resorting to ray shooting'%q)
        return AH_to_V(pp.to_AH_polytope(Z))
   
def to_V(P,N=500):
    r"""
    returns the vertices of the polytopic object $P$ in a vertical stack form.
    """
    if type(P).__name__=='zonotope':
        return zonotope_to_V(P)
    elif type(P).__name__=='AH_polytope':
        return AH_to_V(P,N=N)
    elif type(P).__name__=='H_polytope':
        return H_to_V(P)
    elif type(P).__name__=="hyperbox":
        return zonotope_to_V(P.zonotope)
    else:
        raise ValueError("Did not recognize the polytopic object"+str(type(P).__name__))
    

        
        
def AH_to_H_old(Q,P0,solver="Gurobi"):
    r"""
    Converting Q to an H-polytope using an optimization-based method
    
    WARNING: To be deprecated
    """
    P={}
    P[0]=P0
    if solver=="Gurobi":
        solver=gurobi_solver
    elif solver=="SCS":
        solver=scs_solver
    else:
        raise NotImplementedError
    def find_lambda(P):
        # First solve for Lambda
        program=MP.MathematicalProgram()
        eps=program.NewContinuousVariables(1,"epsilon")
        Ball=pp.hyperbox(N=P.n).H_polytope
        Ball.h=Ball.h*eps
        P_plus_symbolic=pp.minkowski_sum(P[0],Ball)
        program.AddLinearCost(np.array([1]),np.array([0]),eps)
        pp.subset(program,Q,P_plus_symbolic,N=-1)
        Theta,Lambda,Gamma,Beta=pp.subset(program,P[0],Q)
        result=solver.Solve(program,None,None)
        if result.is_success():
            Lambda_n=result.Getsolution(Lambda)
            eps_n=result.Getsolution(eps)
            print("epsilon is",eps_n)
        else:
            raise ValueError("Not feasible")
        return Lambda_n,eps_n
    Lambda,eps=find_lambda(P0)

     
#V=[np.random.random((2,1)) for i in range(10)]
#P=V_polytope(V)
#Q=to_pp.AH_polytope(P)
#
#B=unitbox(4)
#HB=B.H_polytope
#V=H_to_V(HB)
#Q=to_pp.AH_polytope(V_polytope(V))
#ver=AH_to_V(Q)
        
#G=np.random.random((2,5))
#x=np.random.random((2,1))
#Z=zonotope(x,G)
#V=zonotope_to_V(Z)

def vcube(n):
    r"""
    :math:`2^n \times n` array of vectors of vertices in unit cube in :math:`\mathbb{R^n}`
    """
    from itertools import product 
    v=list(product(*list(zip([-1]*n,[1]*n))))
    return np.array(v)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov  5 18:12:46 2019

@author: sadra
"""
from __future__ import division, absolute_import
import numpy as np
#from pypolycontain import objects
#from pypolycontain import conversions
#from pypolycontain import operations
#from pypolycontain import visualize
#from pypolycontain import containment
#from pypolycontain import *
from .objects import *
from .operations import *
from .conversions import *
from .visualize import visualize
from .containment import *


import warnings
import time
import numpy as np

# Scipy
try:
    import scipy.linalg as spa
except:
    warnings.warn("You don't have scipy package installed. You may get error while using some feautures.")

    
#pycdd    
try:
    from cdd import Polyhedron,Matrix,RepType
except:
    warnings.warn("You don't have CDD package installed. Unable to run cone ray generation.")

# Internal Imports
try:
    import pypolycontain as pp
except:
    warnings.warn("You don't have pypolycontain not properly installed.")
#try:
#    from pypolycontain.conversions import to_AH_polytope
#except:
#    pass
#    warnings.warn("You don't have pypolycontain not properly installed.")

try:
    import pydrake.solvers.mathematicalprogram as MP
    import pydrake.solvers.gurobi as Gurobi_drake
    import pydrake.solvers.osqp as OSQP_drake
    # use Gurobi solver
    global gurobi_solver,OSQP_solver, license
    gurobi_solver=Gurobi_drake.GurobiSolver()
    license = gurobi_solver.AcquireLicense()
except:
    warnings.warn("You don't have pydrake installed properly. Methods that rely on optimization may fail.")
        

"""
***********
Description
***********

The necessary and sufficient conditions for encoding was provided here.

"""


def theta_k(circumbody,k=0,i=0):
    """
    This is from the Section 4 of the paper [Sadraddini and Tedrake, 2020].
    
    Inputs:
    
        * AH-polytope: ``circumbody``
        * int ``N``: the number of columns added to the subspace of ``U``. 
        For more information, look at the paper.
        *Default* is 0.
    
    Outputs:
    
        * 2D numpy array ``Theta``, which is defined in the paper
    """
    circumbody_AH=pp.to_AH_polytope(circumbody)
    H_y=circumbody_AH.P.H
    q_y=H_y.shape[0]
    if k<0:
        return np.eye(q_y)
    Y=circumbody_AH.T
    # First identify K=[H_y'^Y'+  ker(H_y')]
    S_1=np.dot(np.linalg.pinv(H_y.T),Y.T)
    S_2=spa.null_space(H_y.T)
    if S_2.shape[1]>0:
        S=np.hstack((S_1,S_2))
    else:
        S=S_1
    # phiw>=0. Now with the augmentation 
    S_complement=spa.null_space(S.T)   
    circumbody.dim_complement=S_complement.shape[1]
    number_of_columns=min(k,S_complement.shape[1])
    psi=np.hstack((S,S_complement[:,i:number_of_columns+i]))
    p_mat=Matrix(np.hstack((np.zeros((psi.shape[0],1)),psi)))
    p_mat.rep_type = RepType.INEQUALITY
    poly=Polyhedron(p_mat)
    R=np.array(poly.get_generators())
    r=R[:,1:].T
    Theta=np.dot(psi,r)
    return Theta


def be_in_set(program,point,zonotope):
    """
    It forces a point to be a member of a zonotope
    """
    dimension = zonotope.G.shape[1]
    b=program.NewContinuousVariables( dimension,'b')
    program.AddBoundingBoxConstraint(-1,1,b)
    program.AddLinearConstraint( np.equal(point, zonotope.x+np.dot(zonotope.G , b) ,dtype='object').flatten() )
    return b

def member_in_set(program,x,P):
    """
    Inputs:
        @ program: mathematical program
        # x: point
        # P: my polytope
    Adds the constraint ..math::`x in P` to the mathematical program
    """
    P=pp.to_AH_polytope(P)
    x.reshape(len(x),1)
    zeta=program.NewContinuousVariables( P.P.n,1, 'zeta')
    program.AddLinearConstraint(A=P.P.H,lb=-np.inf*np.ones((P.P.h.shape[0],1)),ub=P.P.h,vars=zeta)
    _f=np.equal(np.dot(P.T,zeta),x-P.t,dtype='object')
    program.AddLinearConstraint(_f.flatten())
    
    
    
def subset(program,inbody,circumbody,k=-1,Theta=None,i=0,alpha=None,verbose=False):
    """
    Adds containment property Q1 subset Q2
    
    Inputs:
        * program: a `pydrake` mathematical program
        * inbody: a polytopic object
        * circumbody: a polytopic object
        * N: 
            * **Default**: :math:`-1``. Sufficient as in [Sadraddini and Tedrake, 2020]
            * pick `0` for necessary and sufficient encoding (may be too slow) (2019b)
            * pick any positive number. As the number is smaller, 
            the condition becomes closer to necessity. However, this may be too slow.
    
    Output:
        * No direct output, adds :\math:`inbody \subseteq circumbody` to the model
    """
    if type(inbody).__name__=="zonotope" and type(circumbody).__name__=="zonotope":
        """
        For the case when both inbody and circumbody sets are zonotope:
        """
        from itertools import product
        #Defining Variables
        Gamma=program.NewContinuousVariables( circumbody.G.shape[1], inbody.G.shape[1], 'Gamma')
        Lambda=program.NewContinuousVariables( circumbody.G.shape[1],'Lambda')

        #Defining Constraints
        program.AddLinearConstraint(np.equal(inbody.G,np.dot(circumbody.G,Gamma),dtype='object').flatten())             #inbody_G = circumbody_G * Gamma
        program.AddLinearConstraint(np.equal(circumbody.x - inbody.x ,np.dot(circumbody.G,Lambda),dtype='object').flatten())    #circumbody_x - inbody_x = circumbody_G * Lambda

        Gamma_Lambda = np.concatenate((Gamma,Lambda.reshape(circumbody.G.shape[1],1)),axis=1)
        comb = np.array( list(product([-1, 1], repeat= Gamma_Lambda.shape[1])) ).reshape(-1, Gamma_Lambda.shape[1])
        if alpha=='scalar' or alpha== 'vector':
            comb= np.concatenate( (comb,-1*np.ones((comb.shape[0],1))) , axis=1)
        
        # Managing alpha
        if alpha==None:
            variable = Gamma_Lambda
        elif alpha=='scalar':
            alfa = program.NewContinuousVariables(1,'alpha')
        elif alpha=='vector':
            alfa=program.NewContinuousVariables( circumbody.G.shape[1],'alpha')
            variable = np.concatenate((Gamma_Lambda, alfa.reshape(-1,1)),axis=1)
        else:
            raise ValueError('alpha needs to be \'None\', \'scalaer\', or \'vector\'')

        # infinity norm of matrxi [Gamma,Lambda] <= alfa
        for j in range(Gamma_Lambda.shape[0]):
            program.AddLinearConstraint(
                A= comb,
                lb= -np.inf * np.ones(comb.shape[0]),
                ub= np.ones(comb.shape[0]) if alpha==None else np.zeros(comb.shape[0]),
                vars= variable[j,:] if alpha!='scalar' else np.concatenate((Gamma_Lambda[j,:], alfa ))
            ) 


        #from pydrake.symbolic import abs as exp_abs
        # Gamma_abs = np.array([exp_abs(Gamma[i,j]) for i in range(circumbody.G.shape[1]) for j in range(inbody.G.shape[1])]).reshape(*Gamma.shape)
        # Lambda_abs = np.array([exp_abs(Lambda[i]) for i in range(circumbody.G.shape[1])]).reshape(circumbody.G.shape[1],1)
        # Gamma_lambda_abs = np.concatenate((Gamma_abs,Lambda_abs),axis=1)
        # infnorm_Gamma_lambda_abs = np.sum(Gamma_lambda_abs,axis=1)

        #[program.AddConstraint( infnorm_Gamma_lambda_abs[i] <= 1) for i in range(circumbody.G.shape[1])]
        
        #program.AddBoundingBoxConstraint(-np.inf * np.ones(circumbody.G.shape[1]) , np.ones(circumbody.G.shape[1]), infnorm_Gamma_lambda_abs)
        # program.AddLinearConstraint(
        #     A=np.eye(circumbody.G.shape[1]),
        #     lb= -np.inf * np.ones(circumbody.G.shape[1]),
        #     ub=np.ones(circumbody.G.shape[1]),
        #     vars=infnorm_Gamma_lambda_abs
        # )
        if alpha==None:
            return Lambda, Gamma 
        else:
            return Lambda, Gamma , alfa



    Q1=pp.to_AH_polytope(inbody)
    Q2=pp.to_AH_polytope(circumbody)
    Hx,Hy,hx,hy,X,Y,xbar,ybar=Q1.P.H,Q2.P.H,Q1.P.h,Q2.P.h,Q1.T,Q2.T,Q1.t,Q2.t
    qx,qy,nx,ny=Hx.shape[0],Hy.shape[0],X.shape[1],Y.shape[1]
    if type(Theta)!=type(None):
        if verbose:
            print("theta already computed")
        pass
    elif k<0:
        Theta=np.eye(qy)
        if verbose:
            print("Using Positive Orthant")
    else:
        if verbose:
            print("Computing theta with k=%d"%k)
        Theta=theta_k(circumbody,k,i)
    Lambda=program.NewContinuousVariables(Theta.shape[1],qx,'Lambda')
    Gamma=program.NewContinuousVariables(ny,nx,'Gamma')
    gamma=program.NewContinuousVariables(ny,1,'gamma')
    # Constraints
    program.AddBoundingBoxConstraint(0,np.inf,Lambda) # Lambda Non-Negative
    program.AddLinearConstraint(np.equal(X,np.dot(Y,Gamma),dtype='object').flatten()) #X=YGamma
    program.AddLinearConstraint(np.equal(ybar-xbar,np.dot(Y,gamma),dtype='object').flatten()) 
    program.AddLinearConstraint(np.equal(np.dot(Lambda,Hx),np.dot(Theta.T,np.dot(Hy,Gamma)),dtype='object').flatten()) 
    program.AddLinearConstraint(np.less_equal(np.dot(Lambda,hx),\
          np.dot(Theta.T,hy)+np.dot(Theta.T,np.dot(Hy,gamma)),dtype='object').flatten())
    return Theta,Lambda,Gamma,gamma

def necessity_gap(X,Y,Theta):
    """
    The necessity gap for the encoding using cone defined by Theta
    
    .. warning:
        To be added later
    """
    alpha_0=alpha(X,Y,theta_k(Y,k=0))
    my_alpha= alpha(X,Y,Theta)
    return 1-my_alpha/alpha_0

def necessity_gap_k(X,Y,plot=True):
    print("\n\n","="*75,"\n","="*75,"\n\t\tComputing Necessity Gaps ")
    print("="*75,"\n","="*75)
    print("k\tTheta.shape \t delta(X,Y,C) \t alpha \t Computation Time")
    print("-"*75)
    alpha_0=alpha(X,Y,theta_k(Y,k=0))
    Table={}        
    Z=pp.to_AH_polytope(Y)
    kappa=int(Z.T.shape[1]-Z.T.shape[0])
    for k in range(kappa+1):
        t_0=time.time()
        Theta=theta_k(Y,k)
#        print(k,"theta shape",Theta.shape,Theta)
        a=alpha(X,Y,Theta)
        delta=1-a/alpha_0
        cpu_time=time.time()-t_0
        print(k, "\t",Theta.shape,"\t %0.03f"%delta,"\t\t %0.03f"%a,"\t\t %0.003f"%cpu_time)
        Table[k]=np.array([Theta.shape[1],delta,a,cpu_time])
    if plot:
        import matplotlib.pyplot as plt
        plt.figure()
        fig, ax1 = plt.subplots()
        color = (0.7,0,0)
        ax1.set_xlabel(r'$k$')
        ax1.set_ylabel(r'#Rows in $\Theta_k$', color=color,FontSize=20)
        ax1.plot(range(kappa+1),[Table[k][0] for k in range(kappa+1)],'-',color=color)
        ax1.plot(range(kappa+1),[Table[k][0] for k in range(kappa+1)],'o',color=color)
        ax1.tick_params(axis='y', labelcolor=color)
        
        ax2 = ax1.twinx()  # instantiate a second axes that shares the same x-axis
        
        color = (0,0,0.7)
        ax2.set_xlabel(r'$k$')
        ax2.set_ylabel(r'$\delta(\mathbb{X},\mathbb{Y},\mathbb{C}_k)$', color=color,FontSize=20)
        ax2.plot(range(kappa+1),[Table[k][1] for k in range(kappa+1)],'-',color=color)
        ax2.plot(range(kappa+1),[Table[k][1] for k in range(kappa+1)],'o',color=color)
        ax2.tick_params(axis='y', labelcolor=color)
        
        ax2.set_title(r'$k^*=%d, n=%d$'%(kappa,Z.n),FontSize=20)
        ax1.grid()
        fig.tight_layout()  # otherwise the right y-label is slightly clipped
    
    return Table

    
def alpha(X,Y,Theta):
    X2=pp.to_AH_polytope(X)
    Y2=pp.to_AH_polytope(Y)
    prog=MP.MathematicalProgram()
    alpha=prog.NewContinuousVariables(1,"alpha")
    t=prog.NewContinuousVariables(X2.n,1,"t")
    Y2.P.h=Y2.P.h*alpha
    X2.t=X2.t+t
    subset(prog,X2,Y2,Theta=Theta)
    prog.AddLinearCost(np.eye(1),np.zeros((1)),alpha)
    result=gurobi_solver.Solve(prog,None,None)
    if result.is_success():
#        print("alpha test successful")
#        print(result.GetSolution(alpha),result.GetSolution(t))
        return 1/result.GetSolution(alpha)[0]
    else:
        print("not a subset") 


# Numpy
import numpy as np
from itertools import combinations
from scipy.linalg import block_diag

# Pypolycontain
import pypolycontain as pp

class H_polytope():
    r"""
    An H-polytope is a set defined as follows:

    .. math:: 
        \mathbb{P}=\{x \in \mathbb{R}^n  | H x \le h \},
        
    where 
    
    Attributes:
        * :math:`H \in \mathbb{R}^{q \times n}`:  `numpy.ndarray[float[[q,n]]`
        * :math:`h\in \mathbb{R}^q` : `numpy.ndarray[float[[q,1]]`
    
    define the hyperplanes. The inequality is 
    interpreted element-wise and q is the number of hyperplanes. 
    """
    def __init__(self,H,h,symbolic=False,color='red'):
        # Sanity Checks (H should be q*n matrix and h should be q*1 vector)
#        h=np.atleast_2d(h) # I want the vector to be n*1
        if h.shape[0]==1 or len(h.shape)==1:
            h=h.reshape(len(h),1)
        if h.shape[1]!=1:
            ValueError("Error: not appropriate h size, it is",h.shape)
        if H.shape[0]!=h.shape[0]:
            ValueError("Error: not consistent dimension of H: %d and h: %d"%(H.shape[0],h.shape[0]))
        # End of sanity checks
        self.H,self.h=H,h
        self.type="H_polytope"
#        if type(h[0,0]):
#            self.H,self.h=unique_rows(H,h)
        self.n=H.shape[1]
        self.hash_value = None
        self.distance_program=None
        self.color=color

    def __repr__(self):
        return ("H_polytope in R^%d"%self.n)

    def __hash__(self):
        if self.hash_value is None:
            self.hash_value = hash(str(np.hstack([self.H, self.h])))
        return self.hash_value

    def __add__(self,other):
        return pp.minkowski_sum(self,other) # It is returning the minkowski sum in the form of a AH_Polytope!

    def __rmul__(self,scalar):
        """
        Scaling zonotopes by a scalar. The scalar needs to be an integer or a float.
        """
        return pp.H_polytope(self.H,self.h*scalar)
 
class zonotope():
    r"""
    A Zonotope is a set defined as follows:

    .. math:: 
        \mathbb{Z}=\langle x,G \rangle = \{x + G p | p \in [-1,1]^q \},
        
    where 
    
    Attributes:
        * :math:`G \in \mathbb{R}^{n \times q}`: `numpy.ndarray[float[[n,q]]` is the zonotope generator. 
        * :math:`x\in \mathbb{R}^n`: `numpy.ndarray[float[[n,1]]` is the zonotope centroid. Default is zero vector.
    
    The order
    of the zonotope is defined as :math:`\frac{q}{n}`. 
    """
    def __init__(self,G,x=None,name=None,color="green"):
        self.G=np.array(G)        
        if type(x)==type(None):
            self.x=np.zeros((G.shape[0],1))
        else:
            self.x=np.array(x)
        if name==None:
            self.name="zonotope"
        else:
            self.name=name
        self.color=color
        self.type='zonotope'
        self.hash_value = None
        self.distance_program=None
#        self.color="red"

    def __repr__(self):
        return self.name

    def __hash__(self):
        if self.hash_value is None:
            self.hash_value = hash(str(np.hstack([self.G, self.x])))  # FIXME(Kasraghasemi) : better hashing implementation
        return self.hash_value
    
    def __add__(self,other):
        x = self.x + other.x
        G = np.concatenate([self.G,other.G],axis=1)
        return zonotope(G=G,x=x)

    def __pow__(self,other):
        x = np.concatenate([self.x , other.x])
        G = block_diag(self.G,other.G)
        return zonotope(G=G,x=x)

    def __rmul__(self,other):
        """
        Scaling zonotopes by a scalar. The scalar needs to be an integer or a float.
        """
        assert( type(other)==int or type(other)==float or other.shape == (1,) ), "A zonotope can be multiplied by an int or float"
        x = other * self.x
        G = other * self.G
        return zonotope(G=G,x=x)

    def __mul__(self,other):
        return self.__rmul__(other)

    def volume(self):
        r"""
        Computes the volume of the zonotope in :math:`\mathbb{n}` dimensions. 
        The formula is based on the paper in [Gover2002]_
            
            
            
        .. [Gover2002] 
            Gover, Eugene, and Nishan Krikorian. 
            "Determinants and the volumes of parallelotopes and zonotopes." 
            Linear Algebra and its Applications 433, no. 1 (2010): 28-40.
                
        """
        G=self.G
        S=combinations(range(G.shape[1]),G.shape[0])
        V=0
        for s in S:
            Gs=np.hstack([G[:,i:i+1] for i in s])
            V+=abs(np.linalg.det(Gs))
        return (2**G.shape[0]) *V
    
    def volume_gradient(self):
        G=self.G
        S=combinations(range(G.shape[1]),G.shape[0])
        V_dot=np.zeros(G.shape)
        for s in S:
            Gs=np.hstack([G[:,i:i+1] for i in s])
            D=np.linalg.det(Gs)
            if D!=0:
                adj_Gs=D*np.linalg.inv(Gs)
                X=adj_Gs.T*np.sign(np.linalg.det(Gs))
            else:
#                print("Warning: we have a singular matrix")
                e=np.eye(G.shape[0])*10**(-5)
                adj_Gs=np.linalg.det(Gs+e)*np.linalg.inv(Gs+e)
                X=adj_Gs.T*np.sign(np.linalg.det(Gs+e))
#                X=adj_Gs.T*(np.linalg.det(Gs+e))
            for i in range(len(s)):
                V_dot[:,s[i]]+=X[:,i]
        return V_dot

   
class AH_polytope():
    r"""
    An AH_polytope is an affine transformation of an H-polytope and is defined as:
        
    .. math::
        \mathbb{Q}=\{t+Tx  | x \in \mathbb{R}^p, H x \le h \}
    
    Attributes:
        * P: The underlying H-polytope :math:`P:\{x \in \mathbb{R}^p | Hx \le h\}`
        * T: :math:`\mathbb{R}^{n \times p}` matrix: linear transformation
        * t: :math:`\mathbb{R}^{n}` vector: translation 
    """
    def __init__(self,t,T,P,color='blue'):
        """
        Initilization: T,t,P. X=TP+t
        """
        self.T=T # Matrix n*n_p
        self.t=np.atleast_2d(t) # vector n*1
        self.P=P # Polytope in n_p dimensions
        self.type='AH_polytope'
        self.n=T.shape[0]
        if T.shape[1]!=P.H.shape[1]:
            ValueError("Error: not appropriate T size, it is",T.shape[1],P.n)
        self.method="Gurobi"
        self.hash_value = None
        self.distance_program=None
        self.vertices_2D=None
        self.color=color

    def __repr__(self):
        return "AH_polytope from R^%d to R^%d"%(self.P.n,self.n)

    def __hash__(self):
        if self.hash_value is None:
            self.hash_value = hash(self.P) + hash(str(np.hstack([self.T, self.t])))  # FIXME: better hashing implementation
        return self.hash_value
    
    def copy(self):
        return AH_polytope(self.t,self.T,H_polytope(self.P.H,self.P.h))

    def __add__(self,other):
        from pypolycontain.operations import minkowski_sum
        return minkowski_sum(self,other)
    
    def __rmul__(self,scalar):
        return AH_polytope(self.t,self.T*scalar,self.P)


class V_polytope():
    r"""
    V-polytopes are a convex hull of vertices. 
    
    .. math:: 
        \mathbb{V}= \{ x \in \mathbb{R}^n | x = \sum_{i=1}^N \lambda_i v_i,  \sum_{i=1}^N \lambda_i=1, \lambda_i \ge 0, i=1,\cdots,N \}
    
    where each :math:`v_i, i=1,\cdots,N` is a point (some or all are effectively vertices).  
    
    Attributes:
        * list_of_vertices= `list` of `numpy.ndarray[float[n,1]]`.  
    """
    def __init__(self,list_of_vertices):
        self.list_of_vertices=list_of_vertices
        
        
class hyperbox():
    def __init__(self,N=None,corners=[],d=1):
        """
        returns N-dimensional Box 
        corners=typle of 2 numpy arrays: lower_corner and upper_corner
        """
        if len(corners)==0:
            H=np.vstack((np.eye(N),-np.eye(N)))
            h=d*np.ones((2*N,1))
            self.l=-np.ones((N,1))*d
            self.u=-self.l
        else:
            l,u=corners[0:2]
            assert all(u>=l)
            N=l.shape[0]
            assert N==u.shape[0]
            H=np.vstack((np.eye(N),-np.eye(N)))
            if not all(u>=l):
                raise ValueError("Upper-right corner not totally ordering lower-left corner")
            h=np.vstack((u,-l))
            self.l=l
            self.u=u
        self.H_polytope=H_polytope(H,h,color='cyan')
        self.zonotope=zonotope(x=(self.l+self.u)/2,G=np.diagflat((self.u-self.l)/2),color='cyan')
        self.n=N
        self.color='cyan'



class unitbox():
    r"""
    A unitbox in :math:`\mathbb{R}^n` is :math:`[-1,1]^n`.
    """    
    def __init__(self,N):
        H=np.vstack((np.eye(N),-np.eye(N)))
        h=np.ones((2*N,1))
        self.H_polytope=H_polytope(H,h)
        self.zonotope=zonotope(x=np.zeros((N,1)),G=np.eye(N))
        
             
def box(N=None,d=1,corners=[],color='cyan'):
    """
    returns N-dimensional Box 
    corners=typle of 2 numpy arrays: lower_corner and upper_corner
    """
    if len(corners)==0:
        H=np.vstack((np.eye(N),-np.eye(N)))
        h=d*np.ones((2*N,1))
    else:
        l,u=corners[0:2]
        N=l.shape[0]
        assert N==u.shape[0]
        H=np.vstack((np.eye(N),-np.eye(N)))
        if not all(u>=l):
            raise ValueError("Upper-right corner not totally ordering lower-left corner")
        h=np.vstack((u,-l))
    return H_polytope(H,h,color=color)


from setuptools import setup, find_packages


setup(
    name='pypolycontain',
    author='Sadra Sadraddini',
    description='A python package for polytopic objects, operations, and containment encodings',
    author_email='sadra@mit.edu',
    version='1.3',
    packages=['pypolycontain'], #fix
    long_description='A python package for polytopic objects, operations, and containment encodings' 
    +'\nPlease refer to https://pypolycontain.readthedocs.io/en/latest/ for documentation' ,
#    package_dir={'pypolycontain': 'lib'},
    license='MIT',
    url="https://github.com/sadraddini/pypolycontain",
    zip_safe=False,
)
